from unittest.mock import MagicMock

import mido
import pytest


@pytest.fixture
def mock_xtouch(monkeypatch: pytest.MonkeyPatch) -> None:
    mock_in = MagicMock()
    mock_out = MagicMock()

    def _mock_open_input(device_name: str) -> mido.ports.BaseInput:
        if device_name == "X-TOUCH MINI":
            return mock_in

    def _mock_open_output(device_name: str) -> mido.ports.BaseOutput:
        if device_name == "X-TOUCH MINI":
            return mock_out

    monkeypatch.setattr(mido, "open_input", _mock_open_input)
    monkeypatch.setattr(mido, "open_output", _mock_open_output)

    yield mock_in, mock_out



from typing import (
    Iterable,
    Iterator,
    Mapping,
    TypeVar,
)

import mido
import mido.backends
from psygnal import Signal

T = TypeVar("T")


# just a read-only mapping
class _Map(Mapping[int, T]):
    def __init__(self, data: Mapping[int, T]) -> None:
        self._data = data

    def __getitem__(self, key: int) -> T:
        return self._data[key]

    def __iter__(self) -> Iterator[int]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    def __contains__(self, key: object) -> bool:
        return key in self._data


class Button:
    """A button on a midi device."""

    pressed = Signal()
    released = Signal()

    def __init__(self, note: int, output: mido.ports.BaseOutput, channel: int = 10):
        self._note = note
        self._channel = channel
        self._output = output

    def press(self) -> None:
        """Send a note_on message."""
        msg = mido.Message("note_on", channel=self._channel, note=self._note)
        self._output.send(msg)

    def release(self) -> None:
        """Send a note_off message."""
        msg = mido.Message("note_off", channel=self._channel, note=self._note)
        self._output.send(msg)


class Knob:
    """A knob/slider on a midi device."""

    changed = Signal(float)

    def __init__(self, control: int, output: mido.ports.BaseOutput, channel: int = 10):
        self._control = control
        self._channel = channel
        self._output = output

    def set_value(self, val: float) -> None:
        """Send a control_change message."""
        msg = mido.Message(
            "control_change", channel=self._channel, control=self._control, value=val
        )
        self._output.send(msg)

    def __repr__(self) -> str:
        return f"Knob({self._control!r})"


class Buttons(_Map[Button]):
    """A group of buttons. Allows connecting to any button change."""

    pressed = Signal(str)
    released = Signal(str)

    def __init__(
        self, button_ids: Iterable[int], output: mido.ports.BaseOutput
    ) -> None:
        super().__init__({x: Button(x, output) for x in button_ids})
        # connect to any button press/release
        for k, v in self.items():
            v.pressed.connect(lambda k=k: self.pressed.emit(k))
            v.released.connect(lambda k=k: self.released.emit(k))


class Knobs(_Map[Knob]):
    """A group of knobs. Allows connecting to any knob change."""

    changed = Signal(str, float)

    def __init__(self, knob_ids: Iterable[int], output: mido.ports.BaseOutput) -> None:
        super().__init__({x: Knob(x, output) for x in knob_ids})
        # connect to any knob change
        for k, v in self.items():
            v.changed.connect(lambda value, k=k: self.changed.emit(k, value))


class MidiDevice:
    """Generic midi device.

    Parameters
    ----------
    device_name : str
        The name of the midi device to connect to.
    button_ids : Iterable[int]
        The ids of the buttons on the device. (These correspond to the note numbers.)
    knob_ids : Iterable[int]
        The ids of the knobs on the device. (These correspond to the control numbers.)
    """

    def __init__(
        self,
        device_name: str,
        button_ids: Iterable[int] = (),
        knob_ids: Iterable[int] = (),
    ):
        try:
            self._input: mido.ports.BaseInput = mido.open_input(device_name)
            self._output: mido.ports.BaseOutput = mido.open_output(device_name)
        except OSError as e:  # pragma: no cover
            raise OSError(
                f"Could not open input device {device_name!r}. "
                f"Available device names are: {set(mido.get_input_names())}"
            ) from e
        self._input.callback = self._on_msg

        self.device_name = device_name
        self._buttons = Buttons(button_ids, self._output)
        self._knobs = Knobs(knob_ids, self._output)

    @property
    def knob(self) -> Knobs:
        """The knobs on the device.

        Examples
        --------
        >>> x.knob[1].changed.connect(lambda v: print(f"knob 1 changed to {v}"))
        >>> x.knob.changed.connect(lambda k, v: print(f"knob {k} changed to {v}"))
        """
        return self._knobs

    @property
    def button(self) -> Buttons:
        """The buttons on the device.

        Examples
        --------
        >>> x.button[1].pressed.connect(lambda: print("button 1 pressed"))
        >>> x.button[1].released.connect(lambda: print("button 1 released"))
        >>> x.button.pressed.connect(lambda k: print(f"button {k} pressed"))
        >>> x.button.released.connect(lambda k: print(f"button {k} released"))
        """
        return self._buttons

    def close(self) -> None:
        """Close the midi device."""
        self._input.close()
        self._output.close()

    def _on_msg(self, message: mido.Message) -> None:
        if message.type == "control_change":
            self._knobs[message.control].changed.emit(message.value)
        elif message.type == "note_on":
            self._buttons[message.note].pressed.emit()
        elif message.type == "note_off":
            self._buttons[message.note].released.emit()

    def reset(self) -> None:
        """Set all knobs/sliders to 0 and make sure buttons are unpressed."""
        for knob in self._knobs.values():
            knob.set_value(0)
        for button in self._buttons.values():
            button.release()



"""Midi control for microscopes using pymmcore."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pymmcore-midi")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

from ._core_connect import (
    connect_button_to_property,
    connect_device_to_core,
    connect_knob_to_property,
)
from ._device import Button, Knob, MidiDevice
from ._xtouch import XTouchMini

__all__ = [
    "Button",
    "connect_button_to_property",
    "connect_device_to_core",
    "connect_knob_to_property",
    "Knob",
    "MidiDevice",
    "XTouchMini",
]



from unittest.mock import Mock, call

import mido
import pytest
from pymmcore_plus import CMMCorePlus

from pymmcore_midi import XTouchMini, connect_device_to_core, connect_knob_to_property


def test_core_connect(mock_xtouch) -> None:
    mock_in, mock_out = mock_xtouch
    device = XTouchMini()

    core = CMMCorePlus.instance()
    core.loadSystemConfiguration()
    core.setProperty("Camera", "AllowMultiROI", 1)  # show that it highlights the button

    CONNECTIONS = [
        ("button", 8, "Camera", "AllowMultiROI"),
        ("button", 9, "Camera", "Binning"),
        ("knob", 1, "Camera", "Exposure"),
        ("knob", 2, "Camera", "Gain"),
        ("knob", 9, "Camera", "CCDTemperature"),
        ("button", 22, "Core", "startContinuousSequenceAcquisition"),
    ]

    disconnect = connect_device_to_core(device, core, CONNECTIONS)

    # button value was set during connection
    msg = mido.Message("note_on", channel=10, note=8, velocity=64, time=0)
    assert call(msg) in mock_out.send.call_args_list

    with pytest.warns(match="has no limits"):
        connect_knob_to_property(device.knob[3], core, "Camera", "TestProperty5")

    mock = Mock()
    device.knob[2].changed.connect(mock)
    msg = mido.Message("control_change", channel=10, control=4, value=127, time=0)
    mock_in.callback(msg)

    # turn knob
    device.knob[2].changed.emit(127)
    assert core.getProperty("Camera", "Gain") == "8"

    # click button
    assert core.getProperty("Camera", "Binning") == "1"
    device.button[9].released.emit()
    assert core.getProperty("Camera", "Binning") == "2"

    disconnect()



from unittest.mock import MagicMock

import mido

from pymmcore_midi import XTouchMini


def test_x_touch_output(mock_xtouch) -> None:
    """Test sending messages to the X-Touch Mini."""
    _, mock_out = mock_xtouch
    mini = XTouchMini()
    assert len(mini.knob) == 18
    assert len(mini.button) == 48
    assert repr(mini.knob)
    assert 1 in mini.knob

    buttons = {
        "rewind": 18,
        "fast_forward": 19,
        "loop": 20,
        "stop": 21,
        "play": 22,
        "record": 23,
    }
    for name, note in buttons.items():
        btn = getattr(mini, name)
        btn.press()
        msg = mido.Message("note_on", channel=10, note=note, velocity=64, time=0)
        mock_out.send.assert_called_once_with(msg)
        mock_out.reset_mock()
        btn.release()
        msg = mido.Message("note_off", channel=10, note=note, velocity=64, time=0)
        mock_out.send.assert_called_once_with(msg)
        mock_out.reset_mock()

    knob4 = mini.knob[4]
    knob4.set_value(20)
    msg = mido.Message("control_change", channel=10, control=4, value=20, time=0)
    mock_out.send.assert_called_once_with(msg)

    mini.reset()
    mini.close()


def test_x_touch_intput(mock_xtouch) -> None:
    """Test receiving messages from the X-Touch Mini."""
    mock_in, _ = mock_xtouch
    mini = XTouchMini()
    knob4 = mini.knob[4]

    mock = MagicMock()
    knob4.changed.connect(mock)
    msg = mido.Message("control_change", channel=10, control=4, value=30, time=0)
    mock_in.callback(msg)
    mock.assert_called_once_with(30)
    mock.reset_mock()

    btn3 = mini.button[3]
    btn3.pressed.connect(mock)
    msg = mido.Message("note_on", channel=10, note=3, velocity=64, time=0)
    mock_in.callback(msg)
    mock.assert_called_once()
    mock.reset_mock()

    btn3.released.connect(mock)
    msg = mido.Message("note_off", channel=10, note=3, velocity=64, time=0)
    mock_in.callback(msg)
    mock.assert_called_once()

    mini.close()



from pymmcore_plus import CMMCorePlus
from pymmcore_widgets import ImagePreview, PropertyBrowser
from qtpy.QtWidgets import QApplication

from pymmcore_midi import XTouchMini, connect_device_to_core

app = QApplication([])

device = XTouchMini()

core = CMMCorePlus.instance()
core.loadSystemConfiguration()

# Create a PropertyBrowser widget
browser = PropertyBrowser()
browser.show()

core.setProperty("Camera", "AllowMultiROI", 1)  # show that it highlights the button

CONNECTIONS = [
    ("button", 8, "Camera", "AllowMultiROI"),
    ("button", 9, "Camera", "Binning"),
    ("knob", 1, "Camera", "Exposure"),
    ("knob", 2, "Camera", "Gain"),
    ("knob", 3, "Camera", "TestProperty5"),
    ("knob", 9, "Camera", "CCDTemperature"),
    ("button", 21, "Core", "stopSequenceAcquisition"),
    ("button", 22, "Core", "startContinuousSequenceAcquisition"),
]


preview = ImagePreview(mmcore=core)
preview.show()
connect_device_to_core(device, core, CONNECTIONS)

app.exec_()
device.reset()



from ._device import Button, MidiDevice


class XTouchMini(MidiDevice):
    """X-TOUCH MINI midi controller.

    See https://www.behringer.com/product.html?modelCode=P0B3M

    This device has 8 knobs (each of which is also a button), 16 buttons, and 1 sliders.

    Button Indices:

    Layer A:
    --------
    [ 0] [ 1] [ 2] [ 3] [ 4] [ 5] [ 6] [ 7]   (Top Row of Knobs)
    [ 8] [ 9] [10] [11] [12] [13] [14] [15]   (Top Row of Buttons)
    [16] [17] [18] [19] [20] [21] [22] [23]   (Bottom Row of Buttons)

    Layer B:
    --------
    [24] [25] [26] [27] [28] [29] [30] [31]   (Top Row of Knobs)
    [32] [33] [34] [35] [36] [37] [38] [39]   (Top Row of Buttons)
    [40] [41] [42] [43] [44] [45] [46] [47]   (Bottom Row of Buttons)

    Knob/Slider Indices:

    Layer A:
    --------
    [ 1] [ 2] [ 3] [ 4] [ 5] [ 6] [ 7] [ 8]   (Knobs)
    [ 9]                                      (Slider)

    Layer B:
    --------
    [10]                                      (Slider)
    [11] [12] [13] [14] [15] [16] [17] [18]   (Knobs)
    """

    def __init__(self) -> None:
        super().__init__("X-TOUCH MINI", range(48), range(1, 19))

    @property
    def rewind(self) -> Button:
        return self._buttons[18]

    @property
    def fast_forward(self) -> Button:
        return self._buttons[19]

    @property
    def loop(self) -> Button:
        return self._buttons[20]

    @property
    def stop(self) -> Button:
        return self._buttons[21]

    @property
    def play(self) -> Button:
        return self._buttons[22]

    @property
    def record(self) -> Button:
        return self._buttons[23]



from __future__ import annotations

import contextlib
import warnings
from typing import TYPE_CHECKING, Any, Callable, cast

if TYPE_CHECKING:
    from pymmcore_plus import CMMCorePlus

    from pymmcore_midi import Button, Knob, MidiDevice


def connect_knob_to_property(
    knob: Knob, core: CMMCorePlus, device_label: str, property_name: str
) -> Callable[[], None]:
    """Connect a knob to a property controlled by MMCore.

    Parameters
    ----------
    knob : Knob
        The knob to connect
    core : CMMCorePlus
        The MMCorePlus instance
    device_label : str
        The label of the device that owns the property
    property_name : str
        The name of the property to connect to

    Returns
    -------
    Callable[[], None]
        A function that can be called to disconnect the knob from the property
    """
    if not core.hasPropertyLimits(device_label, property_name):
        warnings.warn(
            f"Property {device_label}.{property_name} has no limits and "
            "cannot be connected to a MIDI knob",
            stacklevel=2,
        )
        return lambda: None

    prop_lower = core.getPropertyLowerLimit(device_label, property_name)
    prop_range = core.getPropertyUpperLimit(device_label, property_name) - prop_lower
    knob_lower = 0
    knob_upper = 127
    knob_range = knob_upper - knob_lower

    def knob2value(value: float) -> float:
        """Convert value from knob range to property range."""
        v = (value - knob_lower) / knob_range * prop_range + prop_lower
        return float(v)

    def value2knob(value: float | str) -> int:
        """Convert value from property range to knob range."""
        out = (float(value) - prop_lower) / prop_range * knob_range + knob_lower
        # Make sure the value is in the range [knob_lower, knob_upper]
        return min(max(int(out), knob_lower), knob_upper)

    # set knob value to the current value of the property
    knob.set_value(value2knob(float(core.getProperty(device_label, property_name))))

    # connect knob change events to update core
    @knob.changed.connect
    def _update_core_value(value: float) -> None:
        core.setProperty(device_label, property_name, knob2value(value))

    # connect core property change events to update knob
    def _update_knob_value(dev: str, prop: str, value: str) -> None:
        if dev == device_label and prop == property_name:
            knob.set_value(value2knob(value))

    core.events.propertyChanged.connect(_update_knob_value)

    def disconnect() -> None:
        knob.changed.disconnect(_update_core_value)
        core.events.propertyChanged.disconnect(_update_knob_value)

    return disconnect


def connect_button_to_property(
    button: Button, core: CMMCorePlus, device_label: str, property_name: str
) -> Callable[[], None]:
    """Connect a button to a property controlled by MMCore.

    Parameters
    ----------
    button : Button
        The button to connect
    core : CMMCorePlus
        The MMCorePlus instance
    device_label : str
        The label of the device that owns the property
    property_name : str
        The name of the property to connect to

    Returns
    -------
    Callable[[], None]
        A function that can be called to disconnect the knob from the property
    """
    # set knob value to the current value of the property
    # .set_value(value2knob(float(core.getProperty(device_label, property_name))))

    allowed = core.getAllowedPropertyValues(device_label, property_name)
    is_bool = set(allowed) == {"0", "1"}
    if not allowed:  # pragma: no cover
        warnings.warn(
            f"Property {device_label}.{property_name} has no allowed values and "
            "cannot be connected to a MIDI button",
            stacklevel=2,
        )
        return lambda: None

    def set_button_state(val: Any) -> None:
        if val in ("0", False, 0):
            button.release()  # ensure the button is not highlighted
        elif is_bool:
            button.press()  # keep the button highlighted

    set_button_state(core.getProperty(device_label, property_name))

    # connect knob change events to update core
    @button.released.connect
    def _update_core_value() -> None:
        current = core.getProperty(device_label, property_name)
        next_val = allowed[(allowed.index(current) + 1) % len(allowed)]
        core.setProperty(device_label, property_name, next_val)
        set_button_state(next_val)

    # connect core property change events to update knob
    def _update_button_value(dev: str, prop: str, value: str) -> None:
        if dev == device_label and prop == property_name:
            set_button_state(value)

    core.events.propertyChanged.connect(_update_button_value)

    def disconnect() -> None:
        button.released.disconnect(_update_core_value)
        core.events.propertyChanged.disconnect(_update_button_value)

    return disconnect


def connect_device_to_core(
    device: MidiDevice, core: CMMCorePlus, connections: list[tuple[str, int, str, str]]
) -> Callable[[], None]:
    disconnecters: list[Callable] = []
    for type_, idx, dev, prop in connections:
        if type_ not in ("button", "knob"):  # pragma: no cover
            raise ValueError(f"Unknown type {type_}")

        midi_obj: Knob | Button = getattr(device, type_)[idx]
        if dev == "Core":
            # special case.... look for core method
            if not hasattr(core, prop):  # pragma: no cover
                raise ValueError(f"MMCore object has no method {prop!r}")
            method = getattr(core, prop)
            if type_ == "button":
                btn = cast("Button", midi_obj)
                btn.pressed.connect(method)
                disconnecters.append(lambda o=btn, m=method: o.pressed.disconnect(m))
            elif type_ == "knob":
                # NOTE: connecting a callback to a knob may be a bad idea
                knb = cast("Knob", midi_obj)
                knb.changed.connect(method)
                disconnecters.append(lambda o=knb, m=method: o.changed.disconnect(m))
        elif type_ == "button":
            d = connect_button_to_property(cast("Button", midi_obj), core, dev, prop)
            disconnecters.append(d)
        elif type_ == "knob":
            d = connect_knob_to_property(cast("Knob", midi_obj), core, dev, prop)
            disconnecters.append(d)

    def disconnect() -> None:
        for d in disconnecters:
            with contextlib.suppress(Exception):
                d()

    return disconnect
